#include "Parameter.h"

namespace mdd {
    Parameter::Parameter(IModule* parent)
        : ModuleBase(R"JSON(
        [{
            "name":"appendix",
            "value":0
        }])JSON")
    {
        _parent = parent;
        std::vector<double> default_val = { 1 };
        inputs.push_back(std::make_shared<Input>(parent, "Value", 0, default_val));
        outputs.push_back(std::make_shared<Output>(parent, "Value", 0, default_val));
        
        key = "Parameter";
        setName(key);
    }

    state Parameter::update() {
        return getOutput(0)->setValue(getInput(0)->getValue());
    }

    std::string Parameter::setName(const std::string& name) {
        inputs[0]->setName(name);
        outputs[0]->setName(name);
        return ModuleBase::setName(name);
    }

    std::string Parameter::setAppendix(int appendix)
    {
        inputs[0]->setAppendix(appendix);
        outputs[0]->setAppendix(appendix);
        return ModuleBase::setAppendix(appendix);
    }

    bool Parameter::configureChild(const json& config)
    {
        auto jit = config.find("name");
        if (jit != config.end())
        {
            if (jit.value().get<std::string>() == "appendix")
            {
                setAppendix(jit.value()["value"].get<int>());
                _base_config["configure"]["appendix"]["value"] = getAppendix();
            }
        }
        return false;
    }
}