#include "ProcessorBase.h"

namespace mdd{
    ProcessorBase::ProcessorBase(const std::string& base_config)
    {
        json jparse = json::parse(base_config);
        if (jparse.contains("configure"))
        {
            _base_config = base_config;
        }
        else {
            _base_config["configure"] = jparse;
        }
    }

    bool ProcessorBase::configure(const std::string& config) {
        json jconfig = json::parse(config);
        bool sucess = false;
        auto jit = jconfig.find("GUI");
        if (jit != jconfig.end())
        {
            _base_config["GUI"] = jit.value();
        }
        jit = jconfig.find("params");
        if (jit != jconfig.end())
        {
            
            auto jit2 = jit.value().find("inputs");
            if (jit2 != jit.value().end())
            {
                for (size_t i = 0; i < jit.value().size(); i++)
                {
                    inputs.push_back(std::make_shared<Parameter>(this));
                    inputs.back()->setName(jit2.value()["type"].get<std::string>());
                    inputs.back()->setAppendix(jit2.value()["appendix"].get<int>());
                }
                sucess = true;
            }

            jit2 = jit.value().find("outputs");
            if (jit2 != jit.value().end())
            {
                for (size_t i = 0; i < jit.value().size(); i++)
                {
                    outputs.push_back(std::make_shared<Parameter>(this));
                    outputs.back()->setName(jit2.value()["type"].get<std::string>());
                    outputs.back()->setAppendix(jit2.value()["appendix"].get<int>());
                }
                sucess = true;
            }
        }

        jit = jconfig.find("modules");
        if (jit != jconfig.end()) {
            auto regi = Registration();
            for (size_t i = 0; i < jit.value().size(); i++)
            {
                const json& jmodule = jit.value()[i];
                addModule(regi.generateModule(jmodule["id"].get<std::string>()));
                modules.back()->setName(jmodule["type"].get<std::string>());
                modules.back()->setAppendix(jmodule["appendix"].get<int>());
                modules.back()->load(jmodule["load"]);
            }
            sucess = true;
        }

        jit = jconfig.find("connections");
        if (jit != jconfig.end()) {
            for (size_t i = 0; i < jit.value().size(); i++)
            {
                const json& jconn = jit.value()[i];
                connect(jconn["output"].get<std::string>(), jconn["inputs"].get<std::vector<std::string>>());
            }
            sucess = true;
        }

        jit = jconfig.find("configuration");
        if (jit != jconfig.end()) {
            sucess = configureChild(jit.value());
        }
        return sucess;
    }

    bool ProcessorBase::configureChild(const json& config) {
        return true;
    }

    const std::string& ProcessorBase::getConfiguration() {
        //update module apendix ?
        return _base_config.dump();
    }

    std::string ProcessorBase::getType() {
        return  type;
    }

    std::string ProcessorBase::getGeneratorKey() {
        return  key;
    }

    std::string ProcessorBase::setName(const std::string& name) {
        _name = name;
        return getID();
    }

    std::string ProcessorBase::getName() {
        return  _name;
    }

    std::string ProcessorBase::setAppendix(int appendix) {
        _appendix = appendix;
        return  getID();
    }

    int ProcessorBase::getAppendix() {
        return _appendix;
    }

    std::string ProcessorBase::getID() {
        return _name + std::to_string(_appendix);
    }


    std::string ProcessorBase::addModule(std::shared_ptr<IModule> module){
        modules.emplace_back(module);
        modules.back()->setAppendix(modules.size()-1);
        return modules.back()->getID();
    }

    void ProcessorBase::removeModule(std::shared_ptr<IModule> module)
    {

        for (auto it = modules.begin(); it != modules.end(); ++it) {
            if ((*it) == module)
            {
                modules.erase(it);
            }
        }
    }

    std::vector<std::shared_ptr<Parameter>>& ProcessorBase::getInputParams() {
        return inputs;
    }

    std::vector<std::shared_ptr<Parameter>>& ProcessorBase::getOutputParams() {
        return outputs;
    }

    size_t ProcessorBase::getNumInputs() {
        return processor_inputs.size() + inputs.size();
    }

    size_t ProcessorBase::getNumOutputs() {
        return processor_outputs.size() + outputs.size();
    }
    std::vector< std::shared_ptr<IModule >>& ProcessorBase::getModules(){
        return modules;
    }

    size_t ProcessorBase::getNumModuls() {
        return modules.size();
    }

    std::shared_ptr<IModule> ProcessorBase::getModule(size_t index) {
        if (index < modules.size())
        {
            return modules[index];
        }
        return nullptr;
    }

    bool ProcessorBase::connect(const std::string& output_id, const std::vector<std::string>& input_ids)
    {
        std::size_t found = output_id.find_first_of("/");
        std::string str_module_id = output_id.substr(0, found);
        std::string str_output_id = output_id.substr( found + 1);
        std::shared_ptr<IOutput> out_ptr;
        if (str_module_id.empty())
        {
            out_ptr = getOutput(str_output_id);
        }
        else {
            for (size_t i = 0; i < modules.size(); i++)
            {
                if (modules[i]->getID() == str_module_id)
                {
                    out_ptr = modules[i]->getOutput(str_output_id);
                    break;
                }
            }
        }

        for (size_t i = 0; i < input_ids.size(); i++)
        {
            found = input_ids[i].find_first_of("/");
            str_module_id = input_ids[i].substr(0, found);
            std::string str_input_id = input_ids[i].substr(found + 1);

            if (str_module_id.empty())
            {
                auto in_ptr = getInput(str_input_id);
                in_ptr->connect(out_ptr);
            }
            else {
                for (size_t i = 0; i < modules.size(); i++)
                {
                    if (modules[i]->getID() == str_module_id)
                    {
                        auto in_ptr = modules[i]->getInput(str_input_id);
                        in_ptr->connect(out_ptr);
                        break;
                    }
                }
            }
        }
        return true;
    }

    void ProcessorBase::disconnect() {
        for (size_t i = 0; i < processor_inputs.size(); i++)
        {
            processor_inputs[i]->disconnect();
        }
        for (size_t i = 0; i < processor_outputs.size(); i++)
        {
            processor_outputs[i]->disconnect();
        }
        for (size_t i = 0; i < inputs.size(); i++)
        {
            inputs[i]->disconnect();
        }
        for (size_t i = 0; i < outputs.size(); i++)
        {
            outputs[i]->disconnect();
        }
    }

    void ProcessorBase::load(const json& j) {
        if (j.contains("ID"))
        {
            setAppendix(j["ID"]["appendix"].get<int>());
            setName(j["ID"]["name"].get<std::string>());
            j["ID"]["type"];

        }
        if (j.contains("key"))
        {

        }
        if (j.contains("type"))
        {

        }
        if (j.contains("configure"))
        {

        }

        if (j.contains("modules"))
        {
            Registration regi = Registration();
            for (json jmodule : j["modules"])
            {
                if (jmodule["type"].get<std::string>() == "module")
                {
                    addModule(regi.generateModule(jmodule["key"]));
                }
                else if (jmodule["type"].get<std::string>() == "processor")
                {
                    addModule(regi.generateProcessor(jmodule["key"]));
                }
                modules.back()->load(jmodule);
                int check = 0;
            }
        }

        if (j.contains("inputs"))
        {
            const size_t start_counter = processor_inputs.size();
            for (size_t i = start_counter; i < j["inputs"].size(); i++)
            {

            }
        }

        if (j.contains("outputs"))
        {
            const size_t start_counter = processor_outputs.size();
            for (size_t i = start_counter; i < j["outputs"].size(); i++)
            {

            }
        }

        if (j.contains("connections"))
        {
            for(json jcon : j["connections"])
            {
                json jout = jcon["output"];
                IModule::Ptr module_out_ptr = getModule(jout["prefix"].back().get<std::string>());
                IOutput::Ptr out_ptr = nullptr;
                std::string id = jout["name"].get<std::string>() + std::to_string(jout["appendix"].get<int>());
                for (size_t i = 0; i < module_out_ptr->getNumOutputs(); i++)
                {
                    auto out = module_out_ptr->getOutput(i);
                    if (out->getID() == id)
                    {
                        out_ptr = out;
                        break;
                    }
                }
                for (json jin : jcon["inputs"])
                {
                    IModule::Ptr module_in_ptr = getModule(jin["prefix"].back().get<std::string>());
                    IInput::Ptr in_ptr = nullptr;
                    std::string id = jin["name"].get<std::string>() + std::to_string(jin["appendix"].get<int>());
                    for (size_t i = 0; i < module_in_ptr->getNumInputs(); i++)
                    {
                        auto in = module_in_ptr->getInput(i);
                        if (in->getID() == id)
                        {
                            in_ptr = in;
                            break;
                        }
                    }
                    if (in_ptr != nullptr && out_ptr != nullptr)
                    {
                        in_ptr->connect(out_ptr);
                    }
                }
            }
        }
    }

    json ProcessorBase::dump() {
        json ret = _base_config;
        ret["ID"] = getIdentifier();
        ret["type"] = type;
        ret["key"] = key;

        ret["params"];
        ret["params"]["inputs"];
        for (size_t i = 0; i < inputs.size(); i++)
        {
            json sub;
            sub["type"] = inputs[i]->getType();
            sub["appendix"] = inputs[i]->getAppendix();
            ret["params"]["inputs"].push_back(sub);
        }

        ret["params"]["outputs"];
        for (size_t i = 0; i < outputs.size(); i++)
        {
            json sub;
            sub["type"] = outputs[i]->getType();
            sub["appendix"] = outputs[i]->getAppendix();
            ret["params"]["outputs"].push_back(sub);
        }

        for (auto& in : processor_inputs)
        {
            ret["inputs"].push_back(in->dump());
        }
        for (auto& in : inputs)
        {
            ret["inputs"].push_back(in->getInput(0)->dump());
        }

        for (auto& out : processor_outputs)
        {
            ret["outputs"].push_back(out->dump());
        }
        for (auto& out : outputs)
        {
            ret["outputs"].push_back(out->getInput(0)->dump());
        }
        for (auto& mod : modules)
        {
            if (mod->getType()=="module")
            {
                ret["modules"].push_back(mod->dump());
            }
            else
            {
                ret["modules"].push_back(mod->dump());
            }
           

            for (size_t j = 0; j < mod->getNumOutputs(); j++)
            {
                auto input_connections = mod->getOutput(j)->getConnections();
                if (!input_connections.empty())
                {
                    json connect;
                    connect["output"] = mod->getOutput(j)->getIdentifier();
                    for (size_t k = 0; k < input_connections.size(); k++)
                    {
                        connect["inputs"].push_back(input_connections[k]->getIdentifier());
                    }
                    ret["connections"].push_back(connect);
                }
            }
        }
        return ret;
    }

    json ProcessorBase::getIdentifier() {
        json jID;
        jID["name"] = _name;
        jID["appendix"] = getAppendix();
        jID["type"] = type;
        jID["prefix"] = std::vector<std::string>();
        return jID;
    }

    std::shared_ptr<IOutput> ProcessorBase::getOutput(size_t index) {
        if (index < processor_outputs.size())
        {
            return processor_outputs[index];
        }
        index -= processor_outputs.size();
        if (index < outputs.size())
        {
            return outputs[index]->getOutput(0);
        }
        return nullptr;
    }

    std::vector<std::shared_ptr<IInput>> ProcessorBase::getOptimizableInputs() {
        std::vector<std::shared_ptr<IInput>> ret;
        for (size_t i = 0; i < processor_inputs.size(); i++)
        {
            if (processor_inputs[i]->isOptimizable())
            {
                ret.push_back(processor_inputs[i]);
            }
        }
        for (size_t i = 0; i < inputs.size(); i++)
        {
            auto list = inputs[i]->getOptimizableInputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        for (size_t i = 0; i < modules.size(); i++)
        {
            auto list = modules[i]->getOptimizableInputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        return ret;
    }

    std::vector<std::shared_ptr<IOutput>> ProcessorBase::getOptimizableOutputs() {
        std::vector<std::shared_ptr<IOutput>> ret;
        for (size_t i = 0; i < processor_outputs.size(); i++)
        {
            if (processor_outputs[i]->isOptimizable())
            {
                ret.push_back(processor_outputs[i]);
            }
        }
        for (size_t i = 0; i < outputs.size(); i++)
        {
            auto list = outputs[i]->getOptimizableOutputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        for (size_t i = 0; i < modules.size(); i++)
        {
            auto list = modules[i]->getOptimizableOutputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        return ret;
    }
   
    std::shared_ptr<IOutput> ProcessorBase::getOutput(const std::string& id){
        for (auto& output : processor_outputs) {
            if(output->getID() == id){
                return output;
            }
        }
        for (auto& output : outputs) {
            if(output->getOutput(0)->getID() == id){
                return output->getOutput(0);
            }
        }
        return nullptr;
    }

    std::shared_ptr<IInput> ProcessorBase::getInput(size_t index) {
        if (index < processor_inputs.size())
        {
            return processor_inputs[index];
        }
        index -= processor_inputs.size();
        if (index < inputs.size())
        {
            return inputs[index]->getInput(0);
        }
        return nullptr;
    }

    std::shared_ptr<IInput> ProcessorBase::getInput(const std::string& id){
        for (auto& input : processor_inputs) {
            if(input->getID() == id){
                return input;
            }
        }
        for (auto& input : inputs) {
            if (input->getInput(0)->getID() == id) {
                return input->getInput(0);
            }
        }
        return nullptr;
    }

    std::shared_ptr<IModule> ProcessorBase::getModule(const std::string& id){
        for (auto& module : modules) {
            if(module->getID() == id){
                return module;
            }
        }
        return nullptr;
    }


}
//*/